
/* prob24.c						*/
/* program for LCAU24 option 't'			*/
/* calculate probabilities related to evolution		*/
/* Harold V. McIntosh, 10 August 1987			*/
/* 29 February 1988 - adapted from LCAU21 to LCAU22	*/
/* 11 March 1989 - adapted from LCAU23 to LCAU24	*/
/* 16 March 1989 - draw diagonal in M panel		*/
/* 18 March 1989 - Options n,N - totalistic statistics	*/
/* 15 September 1989 - entropy calculations		*/
/* 22 December 1991 - Bernstein curve fitting &c.	*/
/* 20 January 1992 - separate out prob2r, probk4.	*/

/* references:							*/
/*								*/
/* W. John Wilbur, David J. Lipman and Shihab A. Shamma		*/
/* On the prediction of local patterns in cellular automata	*/
/* Physica 19D 397-410 (1986)					*/ 
/*								*/
/* Howard A. Gutowitz, Jonathan D. Victor and Bruce W. Knight	*/
/* Local structure theory for cellular automata			*/
/* Physica 28D 18-48 (1987)					*/

/*  Copyright (C) 1987  */
/*  Copyright (C) 1988  */
/*  Copyright (C) 1989  */
/*  Harold V. McIntosh  */

# include <math.h>
# include "lcau24.h"

extern char   arr1[AL];
extern char   trule[TS];
extern char   ascrule[KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern char   binrule[KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern int    cp;					/* critical point          */
extern int    pr;					/* probability, mils       */
extern int    co, dn;					/* connectivity, dimension */
extern int    pr, pl;					/* rule, line prob, mils   */
extern int    bi;					/* bernstein index	   */
extern double mf;					/* mouse increment	   */
extern double sm[TS];					/* selfconsistent nbrhd    */
extern double sn[TS];					/* selfconsistent nbrhd    */
extern double bx[BD];					/* bernstein factor        */
extern double bp[BD];
extern double pv[10];					/* polynomial vector 	   */
extern double fpa, fpb, fpc;				/* fixed points		   */

# define GROW	   12	/* row for graph panels		*/
# define BROW	    3	/* row for bar charts		*/
# define EROW	    1	/* row for evolution synopsis	*/
# define AORG 	    0	/* x-origin  of contour plot    */
# define BORG 	  109	/* x-origin  of 2-block param   */
# define CORG 	  219	/* x-origin  of Bernstein plot  */

# include "prob2r.c"
# include "probk4.c"

/* ---------------------------------------------------- */
/* -            edit the probability screen           - */
/* ---------------------------------------------------- */

edtri() {char c; int i; double smq();

  videomode(COLGRAF);
  videopalette(YELREGR);
  prmenu();

  while (0<1) {
  videocursor(0,0,0); videoputc(':',2);
  if (istot()) {scrstr("  T: "); tscrrul(); scrstr("     LCAU24");}
    else if (isstot()) {scrstr("  S: ");
	sscrrul(0); scrstr(",");
	sscrrul(1); scrstr("    LCAU24");}
    else scrrul();
  videocursor(0,0,36); videoputc('?',2);
  c=kbdin();
  if (c == '\015') break;
  videocursor(0,0,38); videoputc(c,2);
  videocursor(0,0,36); videoputc(' ',0);
  switch (c) {
    case '+': videopalette(WHCYMAG); break;
    case '-': videopalette(YELREGR); break;
    case 'a': asfreq(3); break;
    case 'b': rprob(pl); break;
    case 'B': pl=numin(0); prmenu(); break;
    case 'c': traj(numin(0),8,3); break;
    case 'C': trajy(numin(0),0); break;
    case 'd': trajy(numin(0),2500); break;
    case 'D': trajy(numin(0),50000); break;
    case 'e': pevolve(); break;
    case 'E': plevolve(); break;
    case 'f': ranchay(ascrule,pr,DS); break;
    case 'F': pr=numin(0); prmenu(); break;
    case 'g': lifreq(50,1,2); break;
    case 'G': lifreq(200,1,1); break;
    case '.': lifreq(1,1,2); break;
    case 'h': egraf(CORG,0,100); break;
    case 'i': pidiff(100); break;
    case 'j': pjdiff(100); break;
    case 'k': pkdiff(100); break;
    case 'l': pldiff(100); break;
    case 'L': lesq(50,2); break;
    case 'H': lesq(3,2); break;
    case 'm': moncar(1,2); break;
    case 'M': moncar(50,1); break;
    case 'n': clep('m'); neisum(8*BROW,AL,25); break;
    case 'N': clep('m'); nbrsum(8*BROW,AL,10); break;
    case 'o': co=numin(0); prmenu(); break;
    case 'O': dn=numin(0); prmenu(); break;
    case 'p': ptseq(100); break;
    case 'P': hipro(100,0); break;
    case 's': lifreq(50,3,2); break;
    case 'S': lifreq(50,2,2); break;
    case 'X': bx[bi]=smq(); prmenu(); break;
    case 'I': bi=numin(0); prmenu(); break;
    case 'x': pdiff(100); break;
    case 'y': pdiff3(100); break;
    case 't': twoblockfreq(100); break;
    case '1': clep('m'); oneblfreq(8*BROW,300,48); break;
    case '2': clep('m'); twoblfreq(8*BROW,300,48); break;
    case '3': clep('m'); thrblfreq(8*BROW,300,48); break;
    case '4': clep('m'); foublfreq(8*BROW,300,48); break;
    case '5': clep('m'); fivblfreq(8*BROW,300,48); break;
    case '?': trmenu(); break;
    case '!': prmenu(); break;
    case 'z': clep(kbdin()); break;
    case '/': videomode(COLGRAF); videopalette(YELREGR); break;
    case '*': bcprin(); break;
    case '#': parfix(); break;
    case '&': sctf(100,smq()); break;
    case '$': for (i=0; i<11; i++) sctf(100,0.0999*((double)i)); break;
    case '%': clep('m'); neidif(8*BROW); break;
    case 'K': neicon(); break;

    case MOD: gbmar(bi,1); if (bi>0) bi--; gbmar(bi,2); prmenu(); break;
    case MOU: gbmar(bi,1); if (bi<TS-1) bi++; gbmar(bi,2); prmenu(); break;
    case MOR: bx[bi]+=mf; prmenu(); break;
    case MOL: bx[bi]-=mf; prmenu(); break;
    case MOA: mf*=0.5; prmenu(); break;
    case MOB: mf*=2.0; prmenu(); break;
    case MOX: for (i=0; i<BD; i++) bx[i]=1.0; mf=0.05; prmenu(); break;

    default: break;
    } /* end switch */
  }   /* end while  */
  videopalette(WHCYMAG);
  videomode(T80X25);
}      /* end edtri  */


/* -------------------------------------- */
/* show the evolution of a planar section */
/* -------------------------------------- */

plevolve() {int i, j, i0, i1, i2, i3;
  clep('a');
  gfram(AORG,0,100,3);
  asctobin();
  for (i=0; i<5; i++) for (j=0; j<5; j++) {
          i1=i3=0;
	  for (i0=0; i0<16; i0++)
	  for (i2=0; i2<16; i2++)
    {
    videodot(199-3-i2-20*j,AORG+3+i0+20*i,arr1[ix4(i0,i1,i2,i3)]);
    }
    onegen(AL);
    if (kbdst()) {kbdin(); return;}
  }
}




/* evaluate the nth generation Bernstein polynomial at point p */

double bern(p,n) double p; int n; {
int    i, j, d;
double q, s, pq[BD];
d=8*n-7;
pq[0]=q=1.0-p;
for (i=1; i<=d; i++) pq[i]=1.0;
for (i=1; i<=d; i++) {
  for (j=0; j<i; j++) pq[j]*=q;
  for (j=i; j<=d; j++) pq[j]*=p;
  }
for (i=0, s=0.0; i<=d; i++) s+=bp[i]*pq[i];
return(s);
}

/* evaluate the parameters for two-block probabilities after one generation */

twoblock(z) double *z; {
int    i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9;
int    j, j0, j1;
double w, x, y, p, q, pp;
double a[KK][KK], b[KK], c[KK][KK];

p=z[0]; q=1.0-p; pp=z[1];
a[1][1]=pp; a[1][0]=a[0][1]=p-pp; a[0][0]=1.0-2*p+pp;
b[1]=p; b[0]=q;
for (i=0; i<KK; i++) {for (j=0; j<KK; j++) c[i][j]=0.0;}
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8]-'0';
  j1=ascrule[i1][i2][i3][i4][i5][i6][i7][i8][i9]-'0';
  x=a[i0][i1]*a[i1][i2]*a[i2][i3]*a[i3][i4]*a[i4][i5]*a[i5][i6]
	*a[i6][i7]*a[i7][i8]*a[i8][i9];
  y=b[i1]*b[i2]*b[i3]*b[i4]*b[i5]*b[i6]*b[i7]*b[i8];
  w=0.0; if (y>0.0001) w=x/y;
  c[j0][j1]+=w;
  }
z[0]=c[1][0]+c[1][1]; z[1]=c[1][1];
}

/* graph the 2-block parameter differences to find self-consistent values */
/* nxn points 								  */

twoblockfreq(n) int n; {
int i, j, l;
double op[2], np[2], s, t, u, v;

gfram(AORG,0,100,3);

s=1.0/((double)(n));
op[0]=s;
for (i=1; i<n; i++) {
  op[1]=s;
  for (j=1; j<n; j++) {
    np[1]=op[1]; np[0]=op[0]; twoblock(np);
    u=np[1]-op[1]; v=np[0]-op[0];
    t=u*u+v*v;
    if (t<0.0025) l=0; else {
    if (t<0.0675) {if((i+j)%2) l=0; else l=3;} else {
    if (t<0.3906) l=3; else {
    if (t<0.7500) {if((i+j)%2) l=3; else l=2;} else {
    if (t<1.5000) l=2; else {
    if (t<3.0000) {if((i+j)%2) l=2; else l=1;} else {
    if (t<6.0000) l=1; else {
                  if((i+j)%2) l=1; else l=0;
       }}}}}}}
    videodot(199-i,AORG+j,l);
    if (kbdst()) {kbdin(); return;}
    op[1]+=s;
    }  /* end for-j */
  op[0]+=s;
  }    /* end for-i */
}

/* end */

