
/* rijn24.c						*/
/* program for lcau24.c option 'd'			*/
/* display de Bruijn diagram - 1st, 2nd, & 3rd gen	*/
/* Harold V. McIntosh, 20 February 1988			*/
/* 21 February 1988 - add j0=i0; to Pass 4's		*/
/* 3 March 1988 - adapted from lca21.c to lca22.c	*/
/* 12 March 1989 - adapted from lca22.c to lca23.c	*/
/* 5 April 1989 - promoted further to lcau24.c		*/
/* 26 September 1989 - show evolution of periodic conf	*/
/* 2 October 1989 - cycl24.c incorporated into rijn24.c	*/
/* 20 November 1989 - loop in kwait discarded		*/

/*  Copyright (C) 1987	*/
/*  Copyright (C) 1988	*/
/*  Copyright (C) 1989	*/
/*  Harold V. McIntosh	*/
/*  Gerardo Cisneros S.	*/

#include <lcau24.h>

extern int  arr1[AL];
extern char ascrule[KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern int  binrule[KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern int  ix0, iy0;					/* origin for pen moves */

# define LUR		 1	/* left panel upper left row */
# define LUC		 0	/* left panel upper left col */
# define LLR		17	/* left panel lower right row */
# define LLC		15	/* left panel lower right col */
# define RUR		 1	/* right panel upper left row */
# define RUC		16	/* right panel upper left col */
# define RLR		21	/* right panel lower right row */
# define RLC		39	/* right panel lower right col */

# define CUR		 1	/* cycle panel upper left row */
# define CUC		 0	/* cycle panel upper left col */
# define CLR		24	/* cycle panel lower right row */
# define CLC		39	/* cycle panel lower right col */

# define EROW		23	/* row for sample evolution */

# define JX  0.80  /* x-center, de Bruijn diagram	*/
# define JY  0.55  /* y-center, de Bruijn diagram	*/
# define RI 0.400  /* inner radius, de Bruijn diagram	*/
# define RO 0.425  /* outer radius, de Bruijn diagram	*/
# define NW    24  /* pause after so many lines		*/

int  dd;		/* # points in de Bruijn ring	*/
int  mc, nc, nl;	/* column & line control	*/

/* edit the de Bruijn diagram */
edijn() {char c;

  clijn();
  dd=KK*KK*KK*KK;
  ijn(dd,KK,2);
  while (0<1) {
  videocursor(0,0,36);
  videoputc('?',2);
  c=kbdin();
  if (c=='\015') break;
  videocursor(0,0,38);
  videoputc(c,2);
  videocursor(0,0,36);
  videoputc(' ',2);
  videoscroll(1,0,24,14,0,0);
  videocursor(0,0,0);
  kwait(3);
  switch (c) {
    case 'A': dd=KK*KK*KK*KK; ijn(dd,KK,2); sijn(dd,0,1); break;
    case 'B': dd=KK*KK*KK*KK; ijn(dd,KK,2); sijn(dd,1,1); break;
    case 'C': dd=KK*KK*KK*KK; ijn(dd,KK,2); zijn(dd,0,1); break;
    case 'D': dd=KK*KK*KK*KK; ijn(dd,KK,2); zijn(dd,1,1); break;
    case 'E': dd=KK*KK*KK*KK; ijn(dd,KK,2); zijn(dd,2,1); break;
    case 'F': dd=KK*KK*KK*KK; ijn(dd,KK,2); zijn(dd,3,1); break;
    case 'G': dd=KK*KK*KK*KK; ijn(dd,KK,2); zijn(dd,4,1); break;
    case 'a': kwait(0); printf("Precursors 0*:");  kwait(0); xpass1(0); break;
    case 'b': kwait(0); printf("Precursors 1*:");  kwait(0); xpass1(1); break;
    case 'c': kwait(0); printf("(1,-4) gliders:"); kwait(0); apass1(0); break;
    case 'd': kwait(0); printf("(1,-3) gliders:"); kwait(0); apass1(1); break;
    case 'e': kwait(0); printf("(1,-2) gliders:"); kwait(0); apass1(2); break;
    case 'f': kwait(0); printf("(1,-1) gliders:"); kwait(0); apass1(3); break;
    case 'g': kwait(0); printf("(1,0) static: ");  kwait(0); apass1(4); break;
    case 'h': kwait(0); printf("(1,1) gliders:");  kwait(0); apass1(5); break;
    case 'i': kwait(0); printf("(1,2) gliders:");  kwait(0); apass1(6); break;
    case 'j': kwait(0); printf("(1,3) gliders:");  kwait(0); apass1(7); break;
    case 'k': kwait(0); printf("(1,4) gliders:");  kwait(0); apass1(8); break;

    case 'y':
      videoscroll(LUR,LUC,LLR,LLC,0,0);
      videocursor(0,3,0); printf("length = 1...9 \n");
        c=kbdin()-'0'; if (c>0&&c<10) cycle(c); break;

    case '1': dd=KK*KK*KK*KK; clijn(); ijn(dd,KK,2); break;
    case '2': dd=KK*KK*KK*KK*KK*KK*KK*KK; clijn(); ijn(dd,KK,2); break;
    case '3': dd=KK*KK*KK*KK*KK; clijn(); ijn(dd,KK,2); break;
    case '4': dd=KK*KK*KK*KK*KK*KK; clijn(); ijn(dd,KK,2); break;
    case '+': videopalette(WHCYMAG); break;
    case '-': videopalette(YELREGR); break;
    case '?': case '/': clijn(); break;
    default: break;
    } /* end switch */
  }   /* end while  */
  videopalette(WHCYMAG);
  videomode(T80X25);
}

/* clear edijn screen and set up menu */
clijn() {
  videomode(COLGRAF);
  videopalette(YELREGR);
  videocursor(0,0,0);
  scrrul();
  videocursor(0,3,0);
  printf("de Bruijn ring\n\n");
  printf("ab   -  0*,1*\n");
  printf("AB   -  full n*\n");
  printf("cdefghi - (1,x)\n");
  printf("CDEFG - (all)\n");
  printf("h.l.p - (2,x)\n");
  printf("y - cycles 1-9 \n");
  printf("1234 - diagram\n");
  printf("+-   - pallette\n");
  printf("?/   - clr scrn\n");
  printf("<cr> - exit\n\n");
}

/* edit the Pass 4 and Pass 5 shift displays */
edshf(arry,ix,n,p4,p5)
char *arry;
int  *ix, n;
void (*p4)(), (*p5)(); {
char c;
int  i, j, k;
videocursor(0,EROW-5,0); printf("p: list links");
videocursor(0,EROW-4,0); printf("r: random lnk");
videocursor(0,EROW-3,0); printf("s: scan links");
videocursor(0,EROW-2,0); printf("<cr>: quit   ");
while (0<1) {
  c=kbdin();
  if (c=='\015') break;
  switch (c) {
    case 'p':
      videoscroll(LUR,LUC,LLR,LLC,0,0);
      videoscroll(EROW-5,0,EROW-2,15,0,0);
      videoscroll(EROW,0,EROW+1,40,0,0);
      videocursor(0,0,0); kwait(3); kwait(0); kwait(0);
      (*p4)(arry);
      return;
    case 'r':
      for (j=0; j<100; j++) {
        i=rand();
        for (k=0; k<n; k++) {ix[k]=i%KK; i/=KK;}
        (*p5)(arry,ix);
        }
      break;
    case 's':
      for (i=0; i<n; i++) ix[i]=0;
      do {
        (*p5)(arry,ix);
        if (kbdst()) {kbdin(); break;}
        for (i=n; i>0; i--) {
	  ix[i-1]++;
	  if (ix[i-1]==KK) {ix[i-1]=0; ix[i]++;} else break;
	  }} while (i>0);
      videocursor(0,2,0); printf("           ");
      break;
    default: break;
    }}
videoscroll(EROW-5,0,EROW-2,15,0,0);
videoscroll(EROW,0,EROW+1,40,0,0);
}

/* CYCLE24.C					 */

cycle(cy) int cy; {int i, cp; unsigned int arry[19683];
if (cy>9) return;
videoscroll(CUR,CUC,CLR,CLC,0,0);
videocursor(0,1,0); printf("Cycles length %1d",cy); kwait(0);
for (i=0, cp=1; i<cy; i++) cp*=KK;
ylpass1(cy,cp,arry);
ylpass2(cp,arry);
/* ylpass4(cp,arry); */
ylpass5(cy,cp,arry);
}

/* Pass 1 makes arry[i] equal to the successor of i */

ylpass1(cy,cp,arry) int cy, cp; unsigned int *arry; {int i, j, x;
videocursor(0,2,0); printf(" Pass1\015");
asctobin();
for (j=0; j<cp; j++) {
  x=j; for (i=0; i<cy; i++) {arr1[cy-i-1]=x%KK; x/=KK;}
  onegen(cy);
  x=0; for (i=0; i<cy; i++) x=KK*x+arr1[i]; arry[j]=x;
  }
}

/* Pass 2 flags all links with an incoming arrow */

ylpass2(cp,arry) int cp; unsigned int *arry; {int j, m, x;
videocursor(0,2,0); printf(" Pass2\015");
do {
  m=0;
  for (j=0; j<cp; j++) arry[j]|=0x8000;
  for (j=0; j<cp; j++) {x=arry[j];
   if (x!=0xFFFF) arry[x&0x7FFF]&=0x7FFF;}
  for (j=0; j<cp; j++) {
   if ((arry[j]>0x7FFF) && (arry[j]!=0xFFFF)) {arry[j]=0xFFFF; m=1;}}
  } while (m!=0);
}

/* Pass4 - print loops which remain */

ylpass4(cp,arry) int cp; unsigned int *arry; {
int j, x, y, m;
videocursor(0,2,0); printf(" pass4 \015");
for (j=0; j<cp; j++) {
 x=j; y=arry[j]; arry[j]=0xFFFF; m=0;
 while (y!=0xFFFF) {prf(x,y); x=y; y=arry[x]; arry[x]=0xFFFF; m=1;}
 if (m==1) kwait(0);
 }
}

/* print the link */

prf(cy,j,x) int j, x; {int i, y;
  kwait(1);
  y=j; for (i=0; i<cy; i++) {arr1[i]=y%KK; y/=KK;}
  for (i=0; i<cy; i++) printf("%1d",arr1[i]);
  printf("-");
  y=x; for (i=0; i<cy; i++) {arr1[i]=y%KK; y/=KK;}
  for (i=0; i<cy; i++) printf("%1d",arr1[i]);
  printf("  ");
}

/* show sample of the cycles */
ylpass5(cy,cp,arry) int cy, cp; unsigned int *arry; {
int i, j, k, l, x, y, z;
videocursor(0,2,0); printf(" pass5 \015");
for (k=l=0; k<cp; k++) {
  if (arry[k]==0xFFFF) continue;
  z=x=k; y=arry[k]; arry[k]=0xFFFF;
  while (y!=0xFFFF) {x=y; y=arry[x]; arry[x]=0xFFFF;}
  for (i=0; i<AL; i++) if (i%cy==0) {arr1[AL-i-1]=z%KK; x=z/KK;}
    else {arr1[AL-i-1]=x%KK; x/=KK;}
  asctobin();
  for (j=(l%7+1)*25; j<(l%7+1)*25+21; j++) videodot(j,AL-1,2);
  for (j=(l%7+1)*25; j<(l%7+1)*25+21; j++) {
    for (i=0; i<AL; i++) videodot(j,i,arr1[i]);
    onegen(AL); 
    }
  l++;
  if (kbdst()) {kbdin(); break;}
  }
}

/* generate a de Bruijn diagram with m nodes, n links per node, color l */
ijn(m,n,l) int m, n, l; {
int    i, j;
double ii, jj, nn, x, y, t, h;
double sin(), cos();
nn=(double)(n);
t=6.28318/((double)(m));
h=0.5*t;
for (i=0; i<m; i++) {
  ii=(double)(i);
  for (j=0; j<n; j++) {
    jj=(double)(j);
    x=JX-RI*sin(ii*t+h);
    y=JY-RI*cos(ii*t+h);
    videoline(x,y,0);
    x=JX-RO*sin((nn*ii+jj)*t+h);
    y=JY-RO*cos((nn*ii+jj)*t+h);
    videoline(x,y,l);
    }
  }
}

/* insert a link into a de Bruijn diagram */
/* m - number of vertices */
/* u - initial vertex     */
/* v - terminal vertex    */
/* l - color of link	  */
lijnk(m,u,v,l) int m, u, v, l; {
double uu, vv, x, y, t, h;
double sin(), cos();
t=6.28318/((double)(m));
uu=t*((double)(u));
vv=t*((double)(v));
h=0.5*t;
x=JX-RI*sin(uu+h);
y=JY-RI*cos(uu+h);
videoline(x,y,0);
x=JX-RO*sin(vv+h);
y=JY-RO*cos(vv+h);
videoline(x,y,l);
}

/* generate a de Bruijn diagram showing antecedents of state k in color l */
sijn(m,k,l) int m, k, l; {
int    i0, i1, i2, i3, i4, i5, i6, i7, i8;
double th, x, y, t, h;
double sin(), cos();
t=6.28318/((double)(m));
h=0.5*t;
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
  if (k==ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8]-'0') {
    th=((double)(((i0*KK+i1)*KK+i2)*KK+i3)*t+h);
    x=JX-RI*sin(th);
    y=JY-RI*cos(th);
    videoline(x,y,0);
    th=((double)(((i1*KK+i2)*KK+i3)*KK+i4)*t+h);
    x=JX-RO*sin(th);
    y=JY-RO*cos(th);
    videoline(x,y,l);
    }	             /* end if    */
  }}}}}}}}} /* end for's */
}	             /* end sijn  */

/* generate a de Bruijn diagram showing the (1,k) links in color l */
zijn(m,k,l) int m, k, l; {
int    i, i0, i1, i2, i3, i4, i5, i6, i7, i8;
double th, x, y, t, h;
double sin(), cos();
t=6.28318/((double)(m));
h=0.5*t;
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
  i=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8]-'0';
  if ((k==0&&i==i0)||(k==1&&i==i1)||(k==2&&i==i2)||(k==3&&i==i3)||(k==4&&i==i4)) {
    th=((double)(((i0*KK+i1)*KK+i2)*KK+i3)*t+h);
    x=JX-RI*sin(th);
    y=JY-RI*cos(th);
    videoline(x,y,0);
    th=((double)(((i1*KK+i2)*KK+i3)*KK+i4)*t+h);
    x=JX-RO*sin(th);
    y=JY-RO*cos(th);
    videoline(x,y,l);
    }	             /* end if    */
  }}}}}}}}} /* end for's */
}	             /* end zijn  */

/* plot graph on video screen 		   */
/* move pen from present position to (x,y) */
/* raised if l=0; color l if l>0	   */
videoline(x,y,l) double x, y; int l; {
int k, ax, ay, dx, dy, di, dj, x1, y1;
/* if (x<0.0) return; if (x>1.0) return */
/* if (y<0.0) return; if (y>1.0) return */
x1=(int)(199.0*(1.0-y));y1=(int)(199.0*x);
dx=x1-ix0;
dy=y1-iy0;
ax=dx>=0?dx:-dx;
ay=dy>=0?dy:-dy;
di=dx>=0?1:-1; if (dx==0) di=0;
dj=dy>=0?1:-1; if (dy==0) dj=0;

if (l>0) {if (ax!=0 || ay!=0) {
  if (ax>=ay) {for (k=0; k<=ax; k++) videodot(ix0+di*k,60+iy0+(k*dy)/ax,l);}
	 else {for (k=0; k<=ay; k++) videodot(ix0+(k*dx)/ay,60+iy0+dj*k,l);}
  } }
ix0=x1;
iy0=y1;
}

/* approximation to sine */
double sin(x) double x; {
if (x<0.0) return(-sin(-x));
while (x>=6.28318) x-=6.28318;
if (x>=3.142) return(-sin(x-3.14159));
if (x>=1.571) return(sin(3.14159-x));
return(x*(1.0-0.166*(x*x-0.05*x*x)));
}

/* approximation to cosine */
double cos(x) double x; {double sin(); return(sin(x+1.57079));}

/* ONE24.C */
/* arry holds 512 bytes */

/* Pass 1a analyzes all the configurations which fulfil (1,-1+l) */
apass1(l) {
char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK];
int  ix[9], m, n;
void apass4(), apass5();
mc=1;
asctobin();
printf(" Pass1a\015");
for (ix[0]=0; ix[0]<KK; ix[0]++) {
for (ix[1]=0; ix[1]<KK; ix[1]++) {
for (ix[2]=0; ix[2]<KK; ix[2]++) {
for (ix[3]=0; ix[3]<KK; ix[3]++) {
for (ix[4]=0; ix[4]<KK; ix[4]++) {
for (ix[5]=0; ix[5]<KK; ix[5]++) {
for (ix[6]=0; ix[6]<KK; ix[6]++) {
for (ix[7]=0; ix[7]<KK; ix[7]++) {
for (ix[8]=0; ix[8]<KK; ix[8]++) {
  switch (l) {
    case 0: m=ix[0]; break;
    case 1: m=ix[1]; break;
    case 2: m=ix[2]; break;
    case 3: m=ix[3]; break;
    case 4: m=ix[4]; break;
    case 5: m=ix[5]; break;
    case 6: m=ix[6]; break;
    case 7: m=ix[7]; break;
    case 8: m=ix[8]; break;
    default: break;}
  n=binrule[ix[0]][ix[1]][ix[2]][ix[3]][ix[4]][ix[5]][ix[6]][ix[7]][ix[8]];
  arry[ix[0]][ix[1]][ix[2]][ix[3]][ix[4]][ix[5]][ix[6]][ix[7]][ix[8]]=n==m?'Y':'N';
  }}}}}}}}}
apass2i(arry);
apass2o(arry);
edshf(arry,ix,9,apass4,apass5);
}

/* Pass 1x analyzes all the configurations mapping into a constant */
xpass1(l) int l; {
char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK];
int  ix[9], n;
void apass4(), apass5();
mc=1;
asctobin();
printf(" Pass1a\015");
for (ix[0]=0; ix[0]<KK; ix[0]++) {
for (ix[1]=0; ix[1]<KK; ix[1]++) {
for (ix[2]=0; ix[2]<KK; ix[2]++) {
for (ix[3]=0; ix[3]<KK; ix[3]++) {
for (ix[4]=0; ix[4]<KK; ix[4]++) {
for (ix[5]=0; ix[5]<KK; ix[5]++) {
for (ix[6]=0; ix[6]<KK; ix[6]++) {
for (ix[7]=0; ix[7]<KK; ix[7]++) {
for (ix[8]=0; ix[8]<KK; ix[8]++) {
  n=binrule[ix[0]][ix[1]][ix[2]][ix[3]][ix[4]][ix[5]][ix[6]][ix[7]][ix[8]];
  arry[ix[0]][ix[1]][ix[2]][ix[3]][ix[4]][ix[5]][ix[6]][ix[7]][ix[8]]=n==l?'Y':'N';
  }}}}}}}}}
apass2i(arry);
apass2o(arry);
edshf(arry,ix,8,apass4,apass5);
}

/* Pass 2i flags all links with an incoming arrow */
/* Pass 2o flags all links with an outgoing arrow */
/* Then pass 3 discards all unflagged links */
/* Passes 2 and 3 alternate until no change is observed */

apass2i(arry) char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8, m;
do {
printf(" Pass2i\015");
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
if ((arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]&0x5F)=='Y')
  {for (m=0; m<KK; m++) arry[i1][i2][i3][i4][i5][i6][i7][i8][m]|=0x20;}
}}}}}}}}} } while (apass3(arry)!=0); }

apass2o(arry) char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8, m;
do {
printf(" Pass2o\015");
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
if ((arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]&0x5F)=='Y')
  {for (m=0; m<KK; m++) arry[m][i0][i1][i2][i3][i4][i5][i6][i7]|=0x20;}
}}}}}}}}} } while (apass3(arry)!=0); }

/* Pass 3 - erase flags, mark survivors, detect changes */

int apass3(arry) char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8, l;
l=0;
printf(" Pass3 \015");
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
  switch (arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]) {
    case 'y': arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]='Y'; break;
    case 'Y': arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]='N'; l=1; break;
    case 'n': case 'N': arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]='N'; break;
    default: break; }
}}}}}}}}}
return l;
}

/* Pass 4 - print loops which remain */
void apass4(arry) char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8;
int j0, j1, j2, j3, j4, j5, j6, j7, j8;
int k, l, m;
printf(" pass4 \015");
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
for (i7=0; i7<KK; i7++) {
for (i8=0; i8<KK; i8++) {
j0=i0; j1=i1; j2=i2; j3=i3; j4=i4; j5=i5; j6=i6; j7=i7; j8=i8;
l=0; m=0;
do {
        if (arry[0][j1][j2][j3][j4][j5][j6][j7][j8]=='Y')
    {arry[0][j1][j2][j3][j4][j5][j6][j7][j8]='y';
    k=j8; j8=j7; j7=j6; j6=j5; j4=j4; j4=j3; j3=j2; j2=j1; j1=0; m=1;}
  else {if (arry[1][j1][j2][j3][j4][j5][j6][j7][j8]=='Y')
    {arry[1][j1][j2][j3][j4][j5][j6][j7][j8]='y';
    k=j8; j8=j7; j7=j6; j6=j5; j5=j4; j4=j3; j3=j2; j2=j1; j1=1; m=1;}
  else {l=1; if (m==1) 
	{j0=j1; j1=j2; j2=j3; j3=j4; j4=j5; j5=j6; j6=j7; j7=j8; j8=k;} }}
  } while (l==0);
l=0; m=0;
do {
        if (arry[j0][j1][j2][j3][j4][j5][j6][j7][0]=='y')
   {aprf(j0,j1,j2,j3,j4,j5,j6,j7,0);
   arry[j0][j1][j2][j3][j4][j5][j6][j7][0]='N';
   j0=j1; j1=j2; j2=j3; j3=j4; j4=j5; j5=j6; j6=j7; j7=j8; j8=0; m=1;}
  else {if (arry[j0][j1][j2][j3][j4][j5][j6][j7][1]=='y')
   {aprf(j0,j1,j2,j3,j4,j5,j6,j7,1);
   arry[j0][j1][j2][j3][j4][j5][j6][j7][1]='N';
   j0=j1; j1=j2; j2=j3; j3=j4; j4=j5; j5=j6; j6=j7; j7=j8; j8=1; m=1;}
  else {l=1;}}
  } while (l==0);
l=0;
do {
        if (arry[j0][j1][j2][j3][j4][j5][j6][j7][0]=='Y')
   {aprf(j0,j1,j2,j3,j4,j5,j6,j7,0);
   arry[j0][j1][j2][j3][j4][j5][j6][j7][0]='N';
   j0=j1; j1=j2; j2=j3; j3=j4; j4=j5; j5=j6; j6=j7; j7=j8; j8=0; m=1;}
  else {if (arry[j0][j1][j2][j3][j4][j5][j6][j7][1]=='Y')
   {aprf(j0,j1,j2,j3,j4,j5,j6,j7,1);
   arry[j0][j1][j2][j3][j4][j5][j6][j7][1]='N';
   j0=j1; j1=j2; j2=j3; j3=j4; j4=j5; j5=j6; j6=j7; j7=j8; j8=1; m=1;}
  else {l=1; if (m==1) kwait(0);}}
  } while (l==0);
}}}}}}}}}
}

/* print one link of the de Bruijn diagram */
aprf(i0,i1,i2,i3,i4,i5,i6,i7,i8) int i0, i1, i2, i3, i4, i5, i6, i7, i8; {
kwait(1);
printf("%1d",i0);
printf("%1d",i1);
printf("%1d",i2);
printf("%1d",i3);
printf("%1d",i4);
printf("%1d",i5);
printf("%1d",i6);
printf("%1d",i7);
printf("-");
printf("%1d",i8);
printf("  ");
lijnk(dd,((i0*KK+i1)*KK+i2)*KK+i3,((i1*KK+i2)*KK+i3)*KK+i4,3);
}

/* pause at the end of a full screen	    */
/* kwait(0) - short or null line	    */
/* kwait(1) - split long lines as necessary */
/* kwait(2) - wait to continue		    */
/* kwait(3) - reset 			    */
kwait(i) int i; {
switch (i) {
  case 0: printf("\n"); nc=0; nl++; break;
  case 1: if (nc==mc) {printf("&\n"); nc=1; nl++;} else nc++; break;
  case 2: nl=NW; break;
  case 3: nc=0; nl=0; break;
  default: break;}
if (nl==NW) {
  nl=1;
  printf(" press any key to continue\015");
  kbdin();
  printf("                          \015");
  videoscroll(1,0,24,14,0,0);
  videocursor(0,0,0);
  printf("\n");
  }
}

/* Pass 5 - show evolution of sample loops */
void apass5(arry,irry)
char arry[KK][KK][KK][KK][KK][KK][KK][KK][KK];
int  irry[9]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8;
int j1, j2, j3, j4, j5, j6, j7;
int b[KK], i, j, k, l;
i0=irry[0]; i1=irry[1]; i2=irry[2]; i3=irry[3];
i4=irry[4]; i5=irry[5]; i6=irry[6]; i7=irry[7]; i8=irry[8];
if (arry[i0][i1][i2][i3][i4][i5][i6][i7][i8]=='N') return;
lijnk(dd,i0*KK+i1,i1*KK+i2,3);
videocursor(0,2,0); printf(" pass5 \015");
arr1[AL/2-4]=i0;
arr1[AL/2-3]=i1;
arr1[AL/2-2]=i2;
arr1[AL/2-1]=i3;
arr1[AL/2]=i4;
arr1[AL/2+1]=i5;
arr1[AL/2+2]=i6;
arr1[AL/2+3]=i7;
arr1[AL/2+4]=i8;
j1=i1; j2=i2; j3=i3; j4=i4; j5=i5; j6=i6; j7=i7;
for (k=AL/2-5; k>=0; k--) {
  for (i=j=0; j<KK; j++) if (arry[j][i0][i1][i2][i3][i4][i5][i6][i7]=='Y') {b[i]=j; i++;}
  if (i>0) l=rand()%i; else return;
  i7=i6; i6=i5; i5=i4; i4=i3; i3=i2; i2=i1; i1=i0; i0=arr1[k]=b[l];
  }
i1=j1; i2=j2; i3=j3; i4=j4; i5=j5; i6=j6; i7=j7;
for (k=AL/2+5; k<AL; k++) {
  for (i=j=0; j<KK; j++) if (arry[i1][i2][i3][i4][i5][i6][i7][i8][j]=='Y') {b[i]=j; i++;}
  if (i>0) l=rand()%i; else return;
  i1=i2; i2=i3; i3=i4; i4=i5; i5=i6; i6=i7; i7=i8; i8=arr1[k]=b[l];
  }
videoscroll(EROW,0,EROW+1,40,0,0);
asctobin();
for (j=8*EROW; j<8*(EROW+2)-2; j++) videodot(j,AL-1,2);
for (j=8*EROW; j<8*(EROW+2)-2; j++) {
  for (i=0; i<AL; i++) videodot(j,i,arr1[i]);
  onegen(AL);
  }
}

/* TW024.C */
/* no space for arry (128K bytes needed) */

/* end */
